﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ScintillaNet;
using System.ComponentModel;
using System.Windows.Forms;
using System.Drawing;
using System.Runtime.InteropServices;

namespace RikaiChanClone
{
    class RikaiChanScintilla : Scintilla
    {
        static FloatWindow floatWindow = null;
        BackgroundWorker backgroundWorker;
        Timer textChangedTimer = new Timer();
        int previousFirstChar = -1;
        int previousLastChar = -1;

        public RikaiChanScintilla()
            : base()
        {
            if (DesignMode == true || LicenseManager.UsageMode == LicenseUsageMode.Designtime)
            {
                return;
            }

            this.FontChanged += new EventHandler(RikaiChanScintilla_FontChanged);

            this.MouseMove += this_MouseMove;
            this.MouseLeave += this_MouseLeave;
            this.TextChanged += new EventHandler<EventArgs>(this_TextChanged);
            this.Scroll += new EventHandler<ScrollEventArgs>(this_Scroll);
            this.SelectionChanged += new EventHandler(this_SelectionChanged);
            this.Resize += new EventHandler(this_Resize);

            this.MouseWheel += new MouseEventHandler(this_MouseWheel);

            textChangedTimer.Tick += new EventHandler(textChangedTimer_Tick);
        }

        void RikaiChanScintilla_FontChanged(object sender, EventArgs e)
        {
            Application.Idle += new EventHandler(ChangeFontWhenIdle);
        }

        void ChangeFontWhenIdle(object sender, EventArgs e)
        {
            if (this.IsDisposed) return;
            var font = this.Font;
            this.ForEachStyle(s =>
            {
                s.Size = font.Size;
                s.Font = font;
            }
            );
            Application.Idle -= ChangeFontWhenIdle;
        }

        void this_MouseWheel(object sender, MouseEventArgs e)
        {
            RescanIfBoundsChange();
        }

        void this_Resize(object sender, EventArgs e)
        {
            RescanIfBoundsChange();
        }

        void this_SelectionChanged(object sender, EventArgs e)
        {
            RescanIfBoundsChange();
        }

        private void RescanIfBoundsChange()
        {
            int firstVisibleCharacter = this.PositionFromPoint(0, 0);
            int lastVisibleCharacter = this.GetCharIndexFromPosition(new Point(this.Width, this.Height));

            if (firstVisibleCharacter == previousFirstChar && previousLastChar == lastVisibleCharacter)
            {
                return;
            }
            previousFirstChar = firstVisibleCharacter;
            previousLastChar = lastVisibleCharacter;
            RescanLater();
        }

        void this_Scroll(object sender, ScrollEventArgs e)
        {
            RescanLater();
        }

        private void RescanLater()
        {
            if (backgroundWorker != null)
            {
                backgroundWorker.CancelAsync();
            }
            textChangedTimer.Stop();
            textChangedTimer.Interval = 125;
            textChangedTimer.Start();
        }

        void textChangedTimer_Tick(object sender, EventArgs e)
        {
            textChangedTimer.Stop();
            HighlightJapaneseText();
        }

        void this_TextChanged(object sender, EventArgs e)
        {
            RescanLater();
        }

        void this_MouseLeave(object sender, EventArgs e)
        {
            if (floatWindow == null) floatWindow = new FloatWindow();
            floatWindow.Hide();
        }

        private void this_MouseMove(object sender, MouseEventArgs e)
        {
            if (floatWindow == null) floatWindow = new FloatWindow();

            var scintilla = (Scintilla)sender;

            int byteIndex = scintilla.GetCharIndexFromPosition(e.Location);

            int textLength = this.GetRange().End;

            bool found = false;

            if (byteIndex >= 0 && byteIndex < textLength)
            {
                int firstStyle = this.Styles.LastPredefined.Index;
                int styleNumber = this.Styles.GetStyleAt(byteIndex);
                if (styleNumber >= firstStyle)
                {
                    found = true;
                    int i;
                    for (i = byteIndex-1; i >= 0; i--)
                    {
                        if (this.Styles.GetStyleAt(i) != styleNumber)
                        {
                            break;
                        }
                    }
                    byteIndex = i + 1;
                }
                else
                {
                    found = false;
                }

                if (found)
                {
                    int maxRemaining = textLength - byteIndex;
                    int max = Math.Min(maxRemaining, 16 * 3);

                    string sub = this.GetRange(byteIndex, byteIndex + max).Text;

                    var lookupText = JEDictionary.LookupAll(sub);

                    if (lookupText == "")
                    {
                        found = false;
                    }
                    else
                    {
                        floatWindow.LabelText = lookupText;
                        var screenLocation = scintilla.PointToScreen(e.Location);
                        screenLocation.Offset(2, 32);
                        floatWindow.Location = screenLocation;
                        NativeMethods.ShowInactiveTopmost(floatWindow);
                    }
                }
            }
            if (!found)
            {
                floatWindow.Hide();
            }
        }

        public static int IndexOfNextThreeByteCharacter(byte[] rawText, int index)
        {
            if (index < 0 || index >= rawText.Length) return -1;
            for (int i = index; i < rawText.Length; i++)
            {
                if ((rawText[i] & 0xE0) == 0xE0)
                {
                    return i;
                }
            }
            return -1;
        }

        class BackgroundWorkerInput
        {
            public int FirstVisibleCharacter;
            public int LastVisibleCharacter;
            public byte[] RawText;
        }

        class BackgroundWorkerOutput
        {
            public List<int> Starts = new List<int>();
            public List<int> Lengths = new List<int>();
            public BackgroundWorkerInput InputArguments;
        }

        private void HighlightJapaneseText()
        {
            //this is required, otherwise calls to get text index from points will fail
            Application.DoEvents();
            
            if (backgroundWorker == null)
            {
                backgroundWorker = new BackgroundWorker();
                backgroundWorker.WorkerSupportsCancellation = true;
                backgroundWorker.DoWork += new DoWorkEventHandler(backgroundWorker_DoWork);
                backgroundWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(backgroundWorker_RunWorkerCompleted);
            }
            if (backgroundWorker.IsBusy)
            {
                backgroundWorker.CancelAsync();
                RescanLater();
            }
            else
            {
                int firstVisibleCharacter = this.PositionFromPoint(0,0);
                int lastVisibleCharacter = this.GetCharIndexFromPosition(new Point(this.Width, this.Height));
                var rawText = this.RawText;

                var input = new BackgroundWorkerInput() { FirstVisibleCharacter = firstVisibleCharacter, LastVisibleCharacter = lastVisibleCharacter, RawText = rawText };
                backgroundWorker.RunWorkerAsync(input);
            }
        }

        void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Cancelled)
            {
                return;
            }

            var result = (BackgroundWorkerOutput)e.Result;
            var starts = result.Starts;
            var lengths = result.Lengths;

            int firstStyle = this.Styles.LastPredefined.Index;
            var rawText = result.InputArguments.RawText;
            int textLength = rawText.Length;

            int styleEven = firstStyle;
            int styleOdd = firstStyle + 1;
            var style1 = this.Styles[styleEven];
            var style2 = this.Styles[styleOdd];

            style1.BackColor = Color.FromArgb(255, 224, 224);
            style2.BackColor = Color.FromArgb(224, 224, 255);

            int firstVisibleCharacter = result.InputArguments.FirstVisibleCharacter;
            int lastVisibleCharacter = result.InputArguments.LastVisibleCharacter;

            int styleToSet = styleEven;
            int evenOrOdd = 0;

            this.GetRange(firstVisibleCharacter, lastVisibleCharacter).SetStyle(this.Styles.Default.Index);

            for (int i = 0; i < starts.Count; i++)
            {
                int start = starts[i];
                int length = lengths[i];

                var newRange = this.GetRange(start, start + length);
                newRange.SetStyle(styleToSet);

                evenOrOdd = (evenOrOdd + 1) & 1;
                if (evenOrOdd == 0) styleToSet = styleEven;
                if (evenOrOdd == 1) styleToSet = styleOdd;
            }            
        }

        public int FindJapaneseWord(int start, out int length, out string text)
        {
            string inputText;
            List<DictionaryEntry> results;
            var rawText = this.RawText;
            int end = rawText.Length;
            int index = FindJapaneseWord(rawText, start, end, out results, out inputText);
            if (results.Count > 0)
            {
                var firstResult = results[0];
                int lengthInCharacters = firstResult.TextLength;
                text = inputText.Substring(0, lengthInCharacters);
                int lengthInBytes = text.GetLengthInBytes();
                length = lengthInBytes;
            }
            else
            {
                length = 0;
                text = "";
            }
            return index;
        }

        public static int FindJapaneseWord(byte[] rawText, int start, int end, out List<DictionaryEntry> results, out string text)
        {
            results = new List<DictionaryEntry>();
            text = "";

            int index = start;

            index = IndexOfNextThreeByteCharacter(rawText, index);
            if (index == -1)
            {
                return -1;
            }
            if (index >= end)
            {
                return -1;
            }

            int textLength = rawText.Length;

            int maxRemaining = textLength - index;
            int max = Math.Min(maxRemaining, 16 * 3);

            text = UTF8Encoding.UTF8.GetString(rawText, index, max);
            results = JEDictionary.Lookup(text);
            return index;
        }

        public static int FindJapaneseWord(byte[] rawText, int start, int end, out int length)
        {
            List<DictionaryEntry> results;
            string text;
            int index= FindJapaneseWord(rawText, start, end, out results, out text);
            if (results.Count > 0)
            {
                var firstResult = results[0];
                int lengthInCharacters = firstResult.TextLength;
                int lengthInBytes = text.Substring(0, lengthInCharacters).GetLengthInBytes();

                length = lengthInBytes;
            }
            else
            {
                length = 0;
            }
            return index;
        }

        public static int FindJapaneseWord(byte[] rawText, int start, out int length)
        {
            return FindJapaneseWord(rawText, start, rawText.Length, out length);
        }

        void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            var worker = (BackgroundWorker)sender;
            var input = (BackgroundWorkerInput)e.Argument;

            int firstVisibleCharacter = input.FirstVisibleCharacter;
            int lastVisibleCharacter = input.LastVisibleCharacter;
            var rawText = input.RawText;

            BackgroundWorkerOutput output = new BackgroundWorkerOutput();
            output.InputArguments = input;
            var starts = output.Starts;
            var lengths = output.Lengths;
            e.Result = output;

            int textLength = rawText.Length;

            int index = firstVisibleCharacter;

            while (true)
            {
                if (worker.CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }
                int wordLength = 0;
                index = FindJapaneseWord(rawText, index, lastVisibleCharacter, out wordLength);
                if (index == -1)
                {
                    break;
                }
                if (wordLength != 0)
                {
                    starts.Add(index);
                    lengths.Add(wordLength);
                    index += wordLength;
                }
                else
                {
                    index++;
                }
            }
        }
    }
    static class ScintillaExtensions
    {
        public static int GetCharIndexFromPosition(this Scintilla scintilla, Point location)
        {
            int y = location.Y;
            int x = location.X;
            return scintilla.PositionFromPoint(x, y);
        }
        public static int GetLengthInBytes(this string str)
        {
            return UTF8Encoding.UTF8.GetByteCount(str);
        }
        public static void ForEachStyle(this Scintilla scintilla, Action<Style> action)
        {
            int max = scintilla.Styles.Max.Index;
            for (int i = 0; i <= max; i++)
            {
                action(scintilla.Styles[i]);
            }
        }
        public static int ToBytePosition(this int charPosition, string text)
        {
            return UTF8Encoding.UTF8.GetByteCount(text.Substring(0, charPosition));
        }
        public static int ToCharPosition(this int bytePosition, byte[] bytes)
        {
            return UTF8Encoding.UTF8.GetCharCount(bytes, 0, bytePosition);
        }
        public static int ToCharPosition(this int bytePosition, string text)
        {
            return ToCharPosition(bytePosition, UTF8Encoding.UTF8.GetBytes(text));
        }
    }
    static class NativeMethods
    {
        private const int SW_SHOWNOACTIVATE = 4;
        private const int HWND_TOPMOST = -1;
        private const uint SWP_NOACTIVATE = 0x0010;

        [DllImport("user32.dll", EntryPoint = "SetWindowPos")]
        static extern bool SetWindowPos(
             int hWnd,           // window handle
             int hWndInsertAfter,    // placement-order handle
             int X,          // horizontal position
             int Y,          // vertical position
             int cx,         // width
             int cy,         // height
             uint uFlags);       // window positioning flags

        [DllImport("user32.dll")]
        static extern bool ShowWindow(IntPtr hWnd, int nCmdShow);

        public static void ShowInactiveTopmost(Form frm)
        {
            ShowWindow(frm.Handle, SW_SHOWNOACTIVATE);
            SetWindowPos(frm.Handle.ToInt32(), HWND_TOPMOST,
            frm.Left, frm.Top, frm.Width, frm.Height,
            SWP_NOACTIVATE);
        }
    }
}
